%% @author Jacob Torrey <torreyji@clarkson.edu>
%% @copyright 2008 Jacob Torrey
%% @doc The main FANG functionality module
-module(fang).
-export([start/0, setup/0, parse/1, save_var/2, get_var/1]).

% Include our defines
-include("defines.hrl").

%% @doc Starts the REPL
%% @spec start() -> none()
start() ->
    setup(),
    get_line().

%% @doc Performs abstraction on saving variables
%% @spec save_var(any(), any()) -> ok
save_var(K, V) ->
    fang_store:save(K, V),
    ok.

%% @doc Returns a saved variable
%% @spec get_var(any()) -> any()
get_var(K) ->
    fang_store:retr(K).

%% @doc Performs debugging routines if needed
%% @spec debug(function()) -> any() | ok
debug(F) ->
    case ?DEBUG of
	true ->
	    F();
	false ->
	    ok
    end.

%% @doc Sets up the initial runtime environment
%% @spec setup() -> ok
setup() ->
    fang_store:start(),
    case filelib:is_file(?BUILTINS) of % If the builtins file exists
	true ->
	    ok;
	false ->
	    fang_builtins:gen_builtins(?BUILTINS) % If not, regenerate it
    end,
    case ?BIGNUM of % Start the Java bignum server if we need to use it
	true ->
	    javabignum:start();
	false ->
	    ok
    end,
    fang_functions:populate(?BUILTINS),
    ok.

%% @doc Performs the top level REPL functionality
%% @spec get_line() -> none()
get_line() ->
    In = io:get_line('> '), % Ask for a line of input
    WONL = string:substr(In, 1, string:len(In) - 1), % Cut off the \n
    print(parse(WONL)), % Make all the magic happen
    io:format("~n", []),
    get_line(). % Repeat

%% @doc Print out the return value in LISP like format
print([]) -> % Prints an empty list
    io:format("()", []);
print(List) when is_list(List) -> % Recursivly prints out a list
    io:format("(", []),
    print(lists:nth(1, List)), % Print out the first element
    P = fun(E) -> io:format(" ", []), print(E) end, % Recursivly prints out each element
    lists:foreach(P, lists:nthtail(1, List)), % Foreach element in the list, except the first one
    io:format(")", []);
print(Singleton) -> % If it is just a single element, print it out
    io:format("~p", [Singleton]).

%% @doc Converts a string to it's respective Erlang format
%% @spec convert(list()) -> (float() | integer() | atom())
convert(Str) ->
    case string:to_float(Str) of % If it smells like a float
	{error, no_float} ->
	    case string:to_integer(Str) of % If it smells like an int
		{error, no_integer} ->
		    list_to_atom(Str); % Must be an atom
		{I, _} ->
		    I
	    end;
	{F, _} ->
	    F
    end.

%% @doc Checks to see if there is a variable, otherwise returns an atom
%% @spec expand(atom()) -> any()
expand(Var) ->
    case get_var({fang, Var}) of
	undefined ->
	    Var;
	V ->
	    V
    end.

%% %% @doc Goes from an Erlang term to a string
%% %% @spec sexp_to_string(any()) -> list()
%% sexp_to_string(List) when is_list(List) ->
%%     F = fun(E, Acc) -> Acc ++ " " ++ sexp_to_string(E) end, % An anonymous function to convert to string per element
%%     "(" ++ unconvert(lists:nth(1, List)) ++ lists:foldl(F, "", lists:nthtail(1, List)) ++ ")"; % Return the first, then folds the previous anonymous function across the rest, appending them together
%% sexp_to_string(Single) ->
%%     unconvert(Single).

%% @doc Converts an Erlang type to a list
%% @spec unconvert(any()) -> list()
unconvert(V) ->
    case is_integer(V) of
	true ->
	    erlang:integer_to_list(V); % If it's an int, convert to a string rep. of that int
	false ->
	    case is_float(V) of
		true ->
		    erlang:float_to_list(V); % If it's a float, convert to a string rep. or that float
		false ->
		    string:strip(erlang:atom_to_list(V), both, $") % Otherwise convert whatever it is to a string
	    end
    end.

%% @doc Maps a function across a deep list
%% @spec deepmap(functio(), list()) -> list()
deepmap(F, L) ->
    deepmap_helper(F, L, []).

%% @doc The helper function for deepmap
%% @spec deepmap_helper(function(), list(), list()) -> list()
deepmap_helper(_, [], Acc) -> % We're all done, return what we've done
    Acc;
deepmap_helper(F, [L | T], Acc) when is_list(L) -> % We've hit another list, recurse!
    deepmap_helper(F, T, Acc ++ [deepmap_helper(F, L, [])]);
deepmap_helper(F, [H | T], Acc) -> % Run the function on the element
    deepmap_helper(F, T, Acc ++ [F(H)]).

%% @doc Takes in a list of arguments, and a form and makes a lambda function and inserts it into the ETS table with auto memoizing
%% @spec lambda(list(), list()) -> atom()
lambda(Args, Form) ->
    LF = fun(InArg) ->
		 case get_var({memo, {Form, InArg}}) of % Check the memoize table
		     undefined -> % Drat, we need to computer it
			 save_var(fang_list, Form),
			 Fun = fun(E, A) -> % Outer loop through the arguments
				       F = fun(E1) -> % Inner loop through the form
					   case E1 == E of % If they match, replace
					       true ->
						   lists:nth(A, InArg);
					       false ->
						   expand(E1)
					   end
					   end,
				       save_var(fang_list, deepmap(F, get_var(fang_list))),
				       A + 1 % Next run through the loop
			       end,
			 lists:foldl(Fun, 1, Args), % Start looping
			 MForm = get_var(fang_list),
			 Ret = eval_sexp(MForm), % Eval the list we've made
			 save_var({memo, {Form, InArg}}, Ret), % Save our return value
			 Ret;
		     Ret -> % We already did it, return
			 Ret
		 end
	 end,
    {A1, A2, A3} = now(), % Return the time
    random:seed(A1, A2, A3), % Seed our PRNG
    Num = random:uniform(1000), % Generate a random tag for the lambda
    save_var(erlang:list_to_atom("lambda_" ++ erlang:integer_to_list(Num)), LF), % Insert the new function
    erlang:list_to_atom("lambda_" ++ erlang:integer_to_list(Num)). % Return the ETS key

%% @doc Takes in a list of arguments, and a form and makes a lambda function and inserts it into the ETS table without auto memoizing
%% @spec dlambda(list(), list()) -> atom()
dlambda(Args, Form) ->
    LF = fun(InArg) ->
		 save_var(fang_list, Form),
		 Fun = fun(E, A) ->
			       F = fun(E1) ->
					   case E1 == E of
					       true ->
						   lists:nth(A, InArg);
					       false ->
						   expand(E1)
					   end
				   end,
			       save_var(fang_list, deepmap(F, get_var(fang_list))),
			       A + 1
		       end,
		 lists:foldl(Fun, 1, Args),
		 MForm = get_var(fang_list),
		 eval_sexp(MForm)
	 end,
    {A1, A2, A3} = now(),
    random:seed(A1, A2, A3),
    Num = random:uniform(1000),
    save_var(erlang:list_to_atom("dlambda_" ++ erlang:integer_to_list(Num)), LF),
    erlang:list_to_atom("dlambda_" ++ erlang:integer_to_list(Num)).

%% @doc Evaluates a SEXP
%% @spec(any()) -> any()
eval_sexp([Func]) when is_integer(Func) ->
    Func; % Numbers return themselves
eval_sexp([Func]) when is_float(Func) ->
    Func; % Floats return themselves
eval_sexp([[quote | [Args]]]) -> % Here's me hitting quote with a sledge hammer
    Args;
eval_sexp([quote | [Arg]]) -> % More pounding
    Arg;
eval_sexp(['if', Cond, T, F]) -> % Selective eval with if :)
    case eval_sexp(Cond) of
	true ->
	    eval_sexp(T);
	false ->
	    eval_sexp(F)
    end;
eval_sexp([length, L]) -> % Returns the length of a list
    io:format("LEN: ~p~n", [L]),
    erlang:length(L);
eval_sexp([print | [Args]]) -> % Prints out a SEXP
    Ret = eval_sexp(Args),
    print(Ret),
    io:format("~n", []),
    Ret;
eval_sexp(['save-state', Filename]) ->
    fang_store:save_state(unconvert(Filename));
eval_sexp(['load-state', Filename]) ->
    fang_store:load_state(unconvert(Filename));
eval_sexp([load, Filename]) ->
    parse_file(unconvert(Filename));
eval_sexp([lambda, Args, Form] = All) -> % Lambda functions, oh boy!
    debug(fun() -> io:format("~p~n", [All]) end),
    lambda(eval_sexp(Args), eval_sexp(Form));
eval_sexp([list | Args]) -> % Makes a list from the arguments
    F = fun(E) -> eval_sexp(E) end,
    lists:map(F, Args);
eval_sexp(['dirty-lambda', Args, Form]) -> % Dirty lambda functions
    dlambda(eval_sexp(Args), eval_sexp(Form));
eval_sexp([Func | Args] = All) -> % Here's where we do the function lookup in ETS
    F = fun(E) -> R = eval_sexp(E), debug(fun() -> io:format("Eval: ~p ~p~n", [E, R]) end), R end,
    EArgs = lists:map(F, Args), % Map eval across each element of the SEXP
    debug(fun() -> io:format("Eval: ~p ~p~n", [All, EArgs]) end),
    case get_var(Func) of % Checks to see if function exists in ETS
	undefined ->
	    [Func | EArgs]; % If not, just return
	Fun -> % We found it, lets use it :)
	    case apply(Fun, [EArgs]) of 
		{'EXIT', E} -> % There was an error, spit it out
		    {error, E};
		R -> % All systems go
		    R
	    end
    end;
eval_sexp(Single) when is_integer(Single) ->
    Single; % Non-list integers are numbers too
eval_sexp(Single) -> 
    expand(Single). % Otherwise return the expansion of the var

%% @doc Parses in a file and executes all the statments in the file
%% @spec parse_file(list()) -> any()
parse_file(Filename) ->
    debug(fun() -> io:format("Loading: ~p~n", [Filename]) end),
    case file:read_file(Filename) of
	{ok, Bin} ->
	    Commands = string:tokens(erlang:binary_to_list(Bin), "\n"),
	    lists:foreach(fun(E) ->
				   parse(E) end, Commands),
	    true;
	_ ->
	    false
    end.
	    
%% @doc Parse in a SEXP
%% @spec parse(list()) -> any()
parse([$( | Line]) -> % The most common one (...)
    Ret = (catch parse(Line, null_func, [])),
    case Ret of 
	{'EXIT', _} ->
	    io:format("Error! Unbalanced parens!~n", []);
	_ ->
	    eval_sexp(Ret)
    end;
parse([$; | _]) ->
    [];
parse([$', $( | Line]) -> % A quoted '(...)
    case (catch parse(Line, quote, [])) of
	[quote | T] ->
	    T;
	{'EXIT', _} ->
	    io:format("Error! Unbalanced parens!~n", [])
    end;
parse([$' | Line]) -> % A quoted '...
    Tokens = string:tokens(Line, " )"),
    Tok = (catch lists:nth(1, Tokens)),
    case Tok of
	{'EXIT', _} ->
	    io:format("Error! Unbalanced quotes!~n", []);
	_ ->
	    convert(Tok)
    end;
parse([$" | Line]) -> % A string "..."
    Tokens = string:tokens(Line, "\""),
    Tok = (catch lists:nth(1, Tokens)),
    case Tok of
	{'EXIT', _} ->
	    io:format("Error! Unbalanced quotes!~n", []);
	_ ->
	    convert("\"" ++ Tok ++ "\"")
    end;
parse(Line) -> % An atom
    eval_sexp(convert(Line)).

%% @doc The parsing helper function
%% @spec parse(list(), atom(), list()) -> any()
parse([$\s | Rest], Func, Args) -> % Spaces are ignored
    parse(Rest, Func, Args); 
parse([$( | Rest], quote, Args) -> % Continue quoting the sub-list
    {LeftOvers, Ret} = parse(Rest, quote, []), % Parse the sub list
    [_ | T] = Ret,
    parse(LeftOvers, quote, Args ++ [T]);
parse([$', $(, $) | Line], Func, Args) -> % Special case for an empty quoted list
    parse(Line, Func, Args ++ [[]]);
parse([$( | Rest], Func, Args) -> % New list
    {LeftOvers, Ret} = parse(Rest, null_func, []),
    parse(LeftOvers, Func, Args ++ [Ret]);
parse([$', $( | Rest], Func, Args) -> % New quoted list
    {LeftOvers, Ret} = parse(Rest, quote, []),
    [quote | T] = Ret,
    parse(LeftOvers, Func, Args ++ [[[quote, T]]]);
parse([$' | Rest], Func, Args) -> % Quoted atoms
    Tokens = string:tokens(Rest, " )"),
    Tok = lists:nth(1, Tokens),
    NewLine = string:substr(Rest, string:len(Tok) + 1),
    parse(NewLine, Func, Args ++ [[quote, convert(Tok)]]);
parse([$)], Func, Args) -> % All done
    [Func] ++ Args;
parse([$) | Rest], Func, Args) -> % The end of a SEXP, on to MOAR
    {Rest, [Func] ++ Args};
parse(Line, null_func, Args) -> % Something else
    Tokens = string:tokens(Line, " )"),
    Tok = lists:nth(1, Tokens),
    NewLine = string:substr(Line, string:len(Tok) + 1),
    parse(NewLine, expand(convert(Tok)), Args); % Expand the list element
parse([$" | Rest], Func, Args) -> % Strings
    Tokens = string:tokens(Rest, "\""),
    Tok = lists:nth(1, Tokens),
    NewLine = string:substr(Rest, string:len(Tok) + 2),
    parse(NewLine, Func, Args ++ [convert("\"" ++ Tok ++ "\"")]);
parse(Line, quote, Args) -> % Quoting an element of a list
    Tokens = string:tokens(Line, " )"),
    Tok = lists:nth(1, Tokens),
    NewLine = string:substr(Line, string:len(Tok) + 1),
    parse(NewLine, quote, Args ++ [convert(Tok)]);
parse(Line, Func, Args) -> % Otherwise, start from scratch
    Tokens = string:tokens(Line, " )"),
    Tok = (catch lists:nth(1, Tokens)),
    NewLine = string:substr(Line, string:len(Tok) + 1),
    parse(NewLine, Func, Args ++ [expand(convert(Tok))]).
