%% @author Jacob Torrey <torreyji@clarkson.edu>
%% @copyright 2008 Jacob Torrey
-module(fang_store).
-export([retr/1, save/2, start/0, stop/0, clearmemo/0, save_state/1, load_state/1]).

%% @todo Make the memoize tree a LRU cache

% The name for the server
-define(REGNAME, cache).

%% @doc Starts the variable cache
%% @spec start() -> ok
start() ->
    register(?REGNAME, spawn(fun() ->
				  loop(gb_trees:empty(), gb_trees:empty()) end)),
    ok.

%% @doc Stops the variable cache
%% @spec stop() -> ok
stop() ->
    ?REGNAME ! stop,
    ok.

%% @doc Saves the current state of the system to a file
%% @spec save_state(list()) -> ok
save_state(Filename) ->
    ?REGNAME ! {save_state, Filename},
    ok.

%% @doc Loads a previously saved state of the system to a file
%% @spec load_state(list()) -> ok
load_state(Filename) ->
    ?REGNAME ! {load_state, Filename},
    ok.

%% @doc Clears the memoize cache
%% @spec clearmemo() -> ok
clearmemo() ->
    ?REGNAME ! clearmemo,
    ok.

%% @doc Saves a value
%% @spec save(any(), any()) -> ok
save({memo, Form}, Value) -> % This is for a memoized cache
    ?REGNAME ! {memosave, Form, Value},
    ok;
save(K, V) ->
    ?REGNAME ! {save, K, V},
    ok.

%% @doc Retreives a value
%% @spec retr(any()) -> any()
retr({memo, K}) ->
    ?REGNAME ! {self(), memoret, K},
    receive
	{memoret, V} ->
	    V
    end;
retr(K) ->
    ?REGNAME ! {self(), ret, K},
    receive
	{ret, V} ->
	    V
    end.

%% @doc A safe save to a GB tree
%% @spec gb_save(gb_tree(), any(), any()) -> gb_tree()
gb_save(Tree, Key, Value) ->
    case gb_trees:is_defined(Key, Tree) of
	true ->
	    gb_trees:update(Key, Value, Tree);
	false ->
	    gb_trees:insert(Key, Value, Tree)
    end.

%% @doc A safe retrevial from a GB tree
%% @spec gb_retr(gb_tree(), any()) -> any()
gb_retr(Tree, K) ->
    case gb_trees:lookup(K, Tree) of
	{value, V} ->
	    V;
	none ->
	    undefined
    end.

%% @doc The storage loop using two gb_trees
%% @spec loop(gb_tree(), gb_tree()) -> none()
loop(GenTree, MemoTree) ->
    receive
	{memosave, K, V} ->
	    loop(GenTree, gb_save(MemoTree, K, V));
	{save, K, V} ->
	    loop(gb_save(GenTree, K, V), MemoTree);
	{Pid, memoret, K} ->
	    Pid ! {memoret, gb_retr(MemoTree, K)},
	    loop(GenTree, MemoTree);
	{Pid, ret, K} ->
	    Pid ! {ret, gb_retr(GenTree, K)},
	    loop(GenTree, MemoTree);
	{save_state, Filename} ->
	    file:write_file(Filename, erlang:term_to_binary(GenTree)),
	    loop(GenTree, MemoTree);
	{load_state, Filename} ->
	    case file:read_file(Filename) of
		{ok, GT} ->
		    loop(erlang:binary_to_term(GT), MemoTree);
		{error, _Err} ->
		    io:format("Error! Can't read file state!~n", []),
		    ok
	    end,
	    loop(GenTree, MemoTree);
	stop ->
	    ok;
	clearmemo ->
	    loop(GenTree, gb_trees:empty())
    end.
